// for copying and licensing see the attached file LICENSE.txt; please keep this file with the software
//

#define VERSION "0.8"

#define _FILE_OFFSET_BITS 64
#define _LARGE_FILES 1
//#define _TIME_BITS 64
//https://sourceware.org/glibc/wiki/YK2038ProofnessDesign

#include <unistd.h>
#include <stdio.h>
#include <malloc.h>
#include <alloca.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <pthread.h>
#include <signal.h>
#include <errno.h>
#include <sys/socket.h>
#include <resolv.h>
#include <netdb.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <sys/types.h>
#include "crt.h"
#include "networking.h"
#include "dane-support.h"


void *usage(bool iserr) {
  FILE *out = iserr ? stderr : stdout;
  fprintf(out,"atea - download via DANE and http[s], in the future up/download also with other protocols,  version " VERSION "\n");
  fprintf(out,"atea tii [-cqhsdn0] [options] [--] url1 [options] [--] url2 ... download (get) the specified urls\n"); 
  fprintf(out,ADDHLP1);
  fprintf(out,"     --cert xy.pem ... manually specify the server certificate\n");
  fprintf(out,"     --dane ... dynamically get the server certificate via dane\n");
  fprintf(out,"     --no-check-cert / -n ... connect via https without checking the certificate\n");
  fprintf(out,"     --[no-]check-time ... do not check certificate timestamp; i.e. allow outdated server cert\n");
  fprintf(out,"     -0 ... default check: dane for https and no check for http\n");
  fprintf(out,"     -6 ... force usage of ipv6\n");
  fprintf(out,"     -h ... print headers (to stderr) / -hh ... print request and headers / -H do not print headers\n");
  fprintf(out,"     -e ... on failure: save error response into output file / -E unset this\n");
  fprintf(out,"     -c / -q ... continue an interrupted download / be quiet\n");
  fprintf(out,"     --to-stdout / -s / --to xy.dat ... print the output to stdout / into xy.dat (overwrites!)\n");
  fprintf(out,ADDHLP2);
  fprintf(out,"atea tuu file1 file2 url1,2 ... upload file1 to url1/file1 (not implemented)\n");
  fprintf(out,"  supported protocols: http, https\n\n");
  exit(iserr?1:0);
}

bool InitCTX(SSL_CTX **ctx) {
  const SSL_METHOD *method;
  OpenSSL_add_all_algorithms();  /* Load cryptos, et.al. */
  SSL_load_error_strings();   /* Bring in and register error messages */
  method = TLS_client_method();  /* Create new client-method instance */
  //method = TLSv1_2_client_method();  /* Create new client-method instance, old deprecated call */
  *ctx = SSL_CTX_new(method);   /* Create new context */
  if ( *ctx == NULL ) { ERR_print_errors_fp(stderr); return false; }
  return true;
}

bool wasa_sigwinch = true, progress_drawed = false, zero_progress_drawed = false; uint64 dld_start_pos, dld_last_drawn_pos = -1; clockid_t clockid = 0;
#define ProgressTraceSize 30
#define isProgrVisible (dld_last_drawn_pos!=-1)
int progressTraceLen, ptIdx; bool have_retried;
struct DldState { struct timespec time; uint64 dld_pos; } progressTrace[ProgressTraceSize];
struct DldState lastProgressState, prevRealProgress;

bool do_continue = false, quiet = false;

void handleSigWinch(int signo) { wasa_sigwinch = true; }

void InitProgress(uint64 dld_pos) {
  int ret;
  dld_start_pos = dld_pos; lastProgressState.dld_pos = dld_pos; progress_drawed = false;
  if(!clockid) {
    clockid = CLOCK_MONOTONIC; ret = clock_gettime(clockid,&lastProgressState.time);
    if(ret) { clockid = CLOCK_REALTIME; clock_gettime(clockid,&lastProgressState.time); }
  } else clock_gettime(clockid,&lastProgressState.time); 
  ptIdx=0; progressTraceLen=1; progressTrace[0] = lastProgressState; prevRealProgress = lastProgressState;
  dld_last_drawn_pos  = -1;
}

void getMeasure( int *value, char *measure, uint64 rawval ) {
  if( rawval < 1000 ) { *value = rawval; *measure = ' '; return; }
  rawval = ( rawval + 500 ) / 1000;
  if( rawval < 1000 ) { *value = rawval; *measure = 'K'; return; }
  rawval = ( rawval + 500 ) / 1000;
  if( rawval < 1000 ) { *value = rawval; *measure = 'M'; return; }
  rawval = ( rawval + 500 ) / 1000;
  if( rawval < 1000 ) { *value = rawval; *measure = 'G'; return; }
  rawval = ( rawval + 500 ) / 1000;
  *value = rawval; *measure = 'T';
}

pthread_mutex_t prnProgressMtx = PTHREAD_MUTEX_INITIALIZER;

void UpdateProgress( uint64 dld_pos, uint64 content_length) {
  int li, ci; dbl_int timediff; int slidersz, predldd, prevdldd, dldd; int prev_termCols; int i; 
  uint64 speed, med_speed; int speed_value; char speed_measure; uint64 mins, hours, days; int pos;
  pthread_mutex_lock(&prnProgressMtx);
  ptIdx++; if(ptIdx>=ProgressTraceSize) ptIdx=0; progressTraceLen = min_macro(progressTraceLen+1,ProgressTraceSize);
  clock_gettime(clockid,&progressTrace[ptIdx].time); progressTrace[ptIdx].dld_pos = dld_pos;
  timediff = get_timediff_ms(&lastProgressState.time,&progressTrace[ptIdx].time);
  //fprintf(stderr,"%llu\n",timediff);
  if( ( timediff < 250 || ( timediff < 1000 && progressTraceLen < ProgressTraceSize && !have_retried ) || content_length <= 0 ) && 
      !( dld_pos == content_length && lastProgressState.dld_pos != dld_pos && ( isProgrVisible || have_retried ) ) ) {
    pthread_mutex_unlock(&prnProgressMtx);
    return;
  }
  if(wasa_sigwinch) { wasa_sigwinch=false; prev_termCols=termCols; queryWinSz(verbosity>=3); if(termCols!=prev_termCols) progress_drawed=false; }
  // len("100% 365d,12:23min 500Gb/s ") = 27
  slidersz = termCols - 27 - 4;
  if( slidersz > 0 ) {
    if(!progress_drawed) {
      fprintf(stderr,"\r |");
      predldd = ( dld_start_pos * slidersz ) / content_length;
      dldd = ( dld_pos * slidersz ) / content_length;
      for(i=0;i<predldd;i++) fputc('+',stderr);
      for(;i<dldd;i++) fputc('=',stderr); 
      for(;i<slidersz;i++) fputc('-',stderr); 
      fprintf(stderr,"| ");
      progress_drawed=true;
    } else {
      prevdldd = ( dld_last_drawn_pos * slidersz ) / content_length;
      dldd = ( dld_pos * slidersz ) / content_length;
      if( prevdldd < dldd ) {
	fprintf(stderr,"\r" e_moveRight "=", prevdldd+2 );
	for(i=prevdldd+1; i<dldd; i++) fputc('=',stderr);
    } }
    lastProgressState = progressTrace[ptIdx];
    dld_last_drawn_pos = dld_pos;
  }
  if( dld_pos > prevRealProgress.dld_pos )
    prevRealProgress = progressTrace[ptIdx];
  pos = slidersz>0 ? slidersz+4 : 2;
  fprintf(stderr,"\r" e_moveRight "%3i%% ", pos, (int)( dld_pos * 100 / content_length ) );
  li = ( ptIdx + ProgressTraceSize - 8 ) % ProgressTraceSize;
  timediff = get_timediff_ms(&progressTrace[li].time,&progressTrace[ptIdx].time);
  if(timediff>0) speed = ( dld_pos - progressTrace[li].dld_pos ) * 8000 / timediff; else speed=-1;

  ci = ( ptIdx + ProgressTraceSize - progressTraceLen + 1 ) % ProgressTraceSize;
  timediff = get_timediff_ms(&progressTrace[ci].time,&progressTrace[ptIdx].time);
  if( timediff > 0 ) {
    med_speed = ( dld_pos - progressTrace[ci].dld_pos ) * 1000 / timediff;
    mins = ( content_length - dld_pos ) / ( med_speed * 60 );
    hours = mins / 60; mins -= hours * 60;
    days = hours / 24; hours -= days * 24;
    if(days>999) fprintf(stderr,">999days ");
    else if(days>0) fprintf(stderr,"%id,%2i:%2imin ", (int)days, (int)hours, (int)mins );
    else fprintf(stderr,"%2i:%2imin ", (int)hours, (int)mins );
  } else fprintf(stderr,"??:??min ");

  if(speed!=-1) {
    getMeasure( &speed_value, &speed_measure, speed );
    if(speed_value>999) fprintf(stderr,">mx%cb/s" e_clearEOL, speed_measure );
    else fprintf(stderr,"%3i%cb/s" e_clearEOL, speed_value, speed_measure ); 
  } else fprintf(stderr,"??b/s" e_clearEOL); 
  fflush(stderr);
  zero_progress_drawed = false;
  pthread_mutex_unlock(&prnProgressMtx);
}

int sd; SSL *ssl = NULL; SSL_CTX *ctx; bool ctx_inited = false; bool wasa_sighup = false;
int64 timeout = 3000;  // given in ms
bool had_timeout = false;

pthread_t progressThread; bool progressThreadStarted = false, viewProgress = false, establishing_conn = false;
pthread_cond_t shallQuitProgress = PTHREAD_COND_INITIALIZER; 

const char *HUPMSG = "received SIGHUP\n";

void handleSigHUP(int signo) { 
  //if(ssl) SSL_shutdown(ssl);   // do not do this here; this is a signal handler!
  if(sd>=0) { close(sd); sd=-2; }
  wasa_sighup = true;
}

void* ProgressThreadExec(void *arg) {
  struct timespec waituntil, curtime; dbl_int timediff; int slidersz, pos, prev_termCols;
  pthread_mutex_lock(&prnProgressMtx);
  while(viewProgress) {
    clock_gettime( CLOCK_REALTIME, &waituntil ); 
    waituntil.tv_nsec += nanoprecision/1000*500; if( waituntil.tv_nsec >= nanoprecision ) { waituntil.tv_sec++; waituntil.tv_nsec -= nanoprecision; }
    pthread_cond_timedwait( &shallQuitProgress, &prnProgressMtx, &waituntil );
    if(!establishing_conn) {
      clock_gettime(clockid,&curtime); 
      if( viewProgress && isProgrVisible && !zero_progress_drawed ) {
	timediff = get_timediff_ms(&lastProgressState.time,&curtime);
	if( timediff > 500 ) {
	  if(wasa_sigwinch) { wasa_sigwinch=false; prev_termCols=termCols; queryWinSz(verbosity>=3); if(termCols!=prev_termCols) progress_drawed=false; }
	  slidersz = termCols - 27 - 4; pos = slidersz>0 ? slidersz+4 : 2; pos += 4; 
	  if( pos + 11 < termCols ) {  // len str = 10 + 1 char against newline
	    fprintf(stderr,"\r" e_moveRight " 0:00, 0b/s" e_clearEOL, pos); fflush(stderr);
	    zero_progress_drawed = true;
	  }
	  lastProgressState.time = curtime;
	}
      } 
      timediff = get_timediff_ms(&prevRealProgress.time,&curtime);
      if( timediff > timeout && timeout > 0 ) {
	if(ssl) SSL_shutdown(ssl);   // this can lead to SIGPIPE - ignoring this signal
	if(sd>=0) { close(sd); sd=-2; }
	had_timeout = true;
	if( verbosity >= 2 ) { 
	  if(isProgrVisible) fputc('\n',stderr); fprintf(stderr,"timeout reached - restarting download\n"); fflush(stderr); 
	  progress_drawed = false; dld_last_drawn_pos = -1; 
	}
	prevRealProgress.time = curtime;
    } }
  }
  pthread_mutex_unlock(&prnProgressMtx);
  return NULL;
}

X509 *required_cert = NULL;  // given by --cert parameter

#define DIGESTBUF_LEN EVP_MAX_MD_SIZE
int fmtCertName( char *buf, int bufsz, const char *hostname, const char *postfix, int i ) {
  if(i<=1) return snprintf(buf,bufsz,"%s%s.pem",hostname,postfix);
  return snprintf(buf,bufsz,"%s%s-%i.pem",hostname,postfix,i);
}

void saveCert( const char *hostname, char *postfix, X509 *cert ) {
  int i = 1, res, bufsz = strlen(hostname)+strlen(postfix)+4+4+1;   // len(".pem") = len("NNNN") = 4
  char filename[bufsz]; BIO *bo; X509 *fileCert = NULL; bool iseq; struct stat statbuf;
  bo = BIO_new(BIO_s_file());
  while( i < 10000 ) {
    fmtCertName(filename,bufsz,hostname,postfix,i);
    res = BIO_read_filename(bo,filename);
    if(res) { 
      PEM_read_bio_X509(bo, &fileCert, 0, NULL);
      iseq = X509_compare_equal(cert,fileCert);
      X509_free(fileCert); fileCert = NULL;
      if(iseq) {
	fprintf(stderr,"server certificate equals the certificate found in '%s'.\n",filename);
	BIO_free(bo); 
	return;
      }
    } else if( lstat(filename,&statbuf) == -1 ) break;
    i++;
  }
  if(i>=10000) fprintf(stderr,"number space for cert file names exhausted.\n");
  else {
    if(!BIO_write_filename(bo,filename)) fprintf(stderr,"error opening '%s' for writing.\n",filename); else {
      PEM_write_bio_X509(bo,cert);
      fprintf(stderr,"server cert written to '%s'.\n",filename);
  } }
  BIO_free(bo); 
}

#define BUFSZ 4096
char buf[BUFSZ];

int printheaders = 0; FILE *hdrout; 
enum FilenameSpec { FS_IMPLICIT = 0, FS_EXPLICIT = 1, FS_STDOUT = 2 } fileNameMode = FS_IMPLICIT;
const char *fileName = NULL; char *dldFileName = NULL; FILE *output; bool overwriteTargetFile = false, saveErrorResponse = false;

bool OpenOutput( struct UrlSpec *urlspec ) {
  int len, maxlen, fd, num = 1;
  if( fileNameMode == FS_STDOUT ) { output = stdout; dldFileName = NULL; return true; }
  if( fileNameMode == FS_IMPLICIT ) fileName = urlspec->target;
  if( fileName == NULL ) { fprintf(stderr,"error: need to specify target filename with --to for %s.\n",urlspec->fullURL); return false; }
  len = strlen(fileName);
  dldFileName = (char*)malloc(maxlen=len+5+13);   // len(".downloading") == 12
  snprintf(dldFileName,maxlen,"%s.downloading",fileName);
  fd = open(dldFileName,O_CREAT|O_EXCL|O_WRONLY,0666);
  if( fd == -1 ) {
    do {
      snprintf(dldFileName,maxlen,"%s-%i.downloading",fileName,num);
      fd = open(dldFileName,O_CREAT|O_EXCL|O_WRONLY,0666);
      if( fd >= 0 ) break;
      num++;
    } while( num < 10000 );
    if( fd == -1 ) { fprintf(stderr,"ran out of filename space: %s - can not download!\n",dldFileName); return false; }
  }
  output = fdopen(fd,"w"); if(!output) { perror(dldFileName); return false; }
  return true;
}

uint64 dld_pos;

bool AppendOutput( struct UrlSpec *urlspec ) {
  DIR *dirs; struct dirent *dirent; bool result = false;
  int len, entlen; struct stat statbuf;
  if( fileNameMode == FS_STDOUT ) { fprintf(stdout,"can not continue download on stdout; downloading from scratch (%s).\n",urlspec->fullURL); return false; }
  if( fileNameMode == FS_IMPLICIT ) fileName = urlspec->target;
  if( fileName == NULL ) return false;
  dirs = opendir("."); len = strlen(fileName); dldFileName = NULL;
  if(!dirs) perror("trying to list directory content:");
  while(true) {
    dirent = readdir(dirs);
    if(!dirent) break;
    entlen = strlen(dirent->d_name);  // len(".downloading") == 12
    if(!strncmp(dirent->d_name,fileName,len) && entlen >= len+12 && !strncmp(dirent->d_name+entlen-12,".downloading",12) && ( entlen == len+12 || dirent->d_name[len] == '-' ) ) {
      if( stat(dirent->d_name,&statbuf) != -1 ) {
	if( statbuf.st_size > dld_pos && (( statbuf.st_mode & S_IFMT ) == S_IFREG ) ) {   // size + is a regular file
	  free(dldFileName); dldFileName = (char*)malloc(entlen+1);
	  memcpy(dldFileName,dirent->d_name,entlen+1);
	  dld_pos = statbuf.st_size;
	  result = true;
    } } }
  }
  closedir(dirs);
  if( result ) {
    if(verbosity>=1) fprintf(stderr,"continuing download of '%s'.\n",dldFileName);
    output = fopen(dldFileName,"a");
    if(!output) { 
      fprintf(stderr,"error opening '%s' for append: ",dldFileName); perror("");
      free(dldFileName); dldFileName=NULL; result = false;
  } }
  return result;
}

void RenameOutput() {
  struct stat statbuf; int num = 1; char *targetName; int maxlen;
  //printf("%s->%s\n",dldFileName,fileName);
  if(!dldFileName) return;
  fclose(output);
  if( overwriteTargetFile || lstat(fileName,&statbuf) == -1 ) {
    if( rename(dldFileName,fileName) != 0 ) { fprintf(stderr,"error renaming %s into ",dldFileName); perror(fileName); }
  } else { // file already exists, find another name
    maxlen = strlen(fileName)+6;
    targetName = (char*)alloca(maxlen);
    do {
      snprintf(targetName,maxlen,"%s.%i",fileName,num);
      if( lstat(targetName,&statbuf) == -1 ) break;
      num++;
    } while( num < 10000 );
    if( num >= 10000 ) fprintf(stderr,"all files exist: %s; not renaming %s.\n",targetName,dldFileName);
    else if( rename(dldFileName,targetName) != 0 ) { fprintf(stderr,"error renaming %s into ",dldFileName); perror(targetName); }
    else fprintf(stderr,"saved as '%s'.\n",targetName);
  }
  free(dldFileName); dldFileName=NULL;
}

bool useSNI = true;

#define InternalErr 11
#define FileErr 12
#define ConnErr 13
#define CertErr 14
#define FileChangedErr 15
#define HttpErr 16

int reconnect( struct UrlSpec *urlspec, void **fd, X509 *required_cert ) {
  X509 *peer_cert; // return: possible errcodes returned by reconnect are reset on retry=0 in mk_HTTP_Request
  establishing_conn = true;
  if(ssl) SSL_shutdown(ssl);  // this can lead to SIGPIPE - ignoring this signal
  if(sd>=0) doclose(sd);
  sd = socket(urlspec->addr->ai_addr->sa_family, SOCK_STREAM, 0);
  if(socket_connect(&sd,urlspec->addr,0)) { establishing_conn=false; return ConnErr; }
  switch(urlspec->protocol) {
    case PR_HTTP:
      // *fd = &sd; - already set
      break;
    case PR_HTTPS:
      if(ssl) SSL_free(ssl); 
      *fd = ssl = SSL_new(ctx);    /* create new SSL connection state */
      SSL_set_fd(ssl,sd);    /* attach the socket descriptor */
      if(useSNI) SSL_set_tlsext_host_name(ssl,urlspec->host);
      if ( SSL_connect(ssl) == -1 ) { ERR_print_errors_fp(stderr); establishing_conn=false; return ConnErr; }
      if(required_cert) {
	peer_cert = SSL_get_peer_certificate(ssl);   /* get the server's certificate */
	if(!X509_compare_equal(peer_cert,required_cert)) { 
	  fprintf(stderr,"reconnect: connected to server with different certificate\n"); X509_free(peer_cert);
	  establishing_conn = false;
	  return CertErr;
	}
	X509_free(peer_cert);
      }
      break;
  }
  establishing_conn = false;
  return 0;
}

int max_retries = 3;

bool mk_HTTP_Request( struct UrlSpec *urlspec, void *fd, PReadIn readin, PWriteOut writeout, X509 *required_cert ) {
  int len, ioc, i, j, k; char *bufptr; bool last_c_nl; char *stopchar; uint64 content_length, cur_content_length; 
  bool had_content_length = false, header_continued, break2, break3; bool headerline; long statuscode; int l; bool saveOutput; int errcode = 0; 
  bool is_continued = false, saveOutput_set = false; bool wasInterrupted = false; int this_errcode; int retry; bool make_retry;

  dld_pos = 0; retry = 0; output = NULL; break3=false; content_length = 0; have_retried = false;
  while( retry <= max_retries || max_retries < 0 ) {
    retry++; make_retry = false; if(retry>1) have_retried = true;
    if(wasInterrupted) {
      this_errcode = reconnect(urlspec,&fd,required_cert);
      if( this_errcode > errcode ) errcode = this_errcode;
      if(this_errcode) continue;
    }
    // assemble request
    len = snprintf(buf,BUFSZ,"GET %s HTTP/1.1\r\n", urlspec->path);
    if(do_continue||wasInterrupted) {
      if(!wasInterrupted) is_continued = AppendOutput(urlspec);
      if(is_continued||wasInterrupted) len += snprintf(buf+len,BUFSZ-len,"Range: bytes=%llu-\r\n",dld_pos); 
    }
    if( urlspec->port_override >= 0 ) 
      if(urlspec->isip6adr) len += snprintf(buf+len,BUFSZ-len,"Host: [%s]:%i",urlspec->host,urlspec->port_override); 
                       else len += snprintf(buf+len,BUFSZ-len,"Host: %s:%i",urlspec->host,urlspec->port_override);
    else
      if(urlspec->isip6adr) len += snprintf(buf+len,BUFSZ-len,"Host: [%s]",urlspec->host);
                       else len += snprintf(buf+len,BUFSZ-len,"Host: %s",urlspec->host);
    len += snprintf(buf+len,BUFSZ-len,"\r\nUser-Agent: atea/" VERSION OSSTR "\r\nAccept: */*\r\nConnection: Keep-Alive\r\nAccept-Encoding: identity\r\n\r\n");
    if( len < 1 || len > BUFSZ-2 ) { fprintf(stderr,"internal error at formatting the request\n"); errcode = InternalErr; break; }
    // send request
    if(!quiet) InitProgress(dld_pos);
    //fwrite(buf,len,1,stdout);
    if(printheaders>=2) fwrite(buf,len,1,hdrout);
    bufptr=buf; while( len > 0 ) {
      ioc = writeout(fd,bufptr,len);
      if( ioc <= 0 ) {
	if(!had_timeout && verbosity >= 2 ) { 
	  if(isProgrVisible) fputc('\n',stderr); fprintf(stderr,wasa_sighup?HUPMSG:"error writing to socket\n"); 
	  dld_last_drawn_pos = -1;
	} 
	had_timeout = false; wasa_sighup = false;
        if( ConnErr > errcode ) errcode = ConnErr;
	make_retry=true; break;
      }
      bufptr+=ioc; len-=ioc;
    }
    if(make_retry) { wasInterrupted=true; continue; }
    bufptr=buf; len=0; header_continued=false; break2=false;
    headerline=true; statuscode=-1;
    // read and parse response header
    while(true) {
      ioc = readin(fd,bufptr,BUFSZ-len);
      if(ioc<0) { 
	if(!had_timeout && verbosity >= 2 ) { 
	  if(isProgrVisible) fputc('\n',stderr); if(wasa_sighup) fprintf(stderr,HUPMSG); else fprintf(stderr,"read error (retry %i)\n",retry); 
	  dld_last_drawn_pos = -1; 
        } 
	had_timeout = false; wasa_sighup = false;
       	if( ConnErr > errcode ) errcode=ConnErr; 
	make_retry=true; break; 
      }
      if(!ioc) { j=len; break;}
      bufptr+=ioc; len+=ioc;
      i=0; last_c_nl = !header_continued; // usually last_c_nl=true
      do {
	j=i; while( j<len && buf[j] != '\n' ) { last_c_nl &= buf[j]=='\r'; j++; }
	if(j>=len) break;
	if(last_c_nl) { j++; break2=true; break; } else last_c_nl=true;
	if(headerline) {  // first line containing status code
	  k=i; while( k<j && buf[k]!=' ' && buf[k]!='\t' ) k++;
	  k++; while( k<j &&(buf[k]==' '||buf[k]=='\t')) k++;
	  l=k+1; while( l<j && buf[l]!='\r' && buf[l]!='\n' ) l++;
	  aux_assert( l<j );  // because of j>=len => break;
	  errno = 0; statuscode = strtol( buf+k, &stopchar, 0 );
	  if( !( errno==0 && ( !*stopchar || *stopchar == '\r' || *stopchar == '\n' || *stopchar == ' ' || *stopchar == '\t' ) ) ) statuscode = -1;
	  if( statuscode!=200 && statuscode!=206 ) {   // 206 ~ partial content 
	    //printheaders = 300 <= statuscode && statuscode < 400;
	    if( !printheaders && statuscode != 416 ) { fwrite(buf+k,l-k,1,stderr); fputc('\n',stderr); }
	    errcode = statuscode/100; if(errcode<2) errcode=2; else if(errcode>9) errcode=2;
	  }
	  headerline=false;
	} else if( !strncasecmp(buf+i,"Content-Length:",min_macro(15,j-i)) && ( !header_continued || i>0 ) ) { 
	  k=i+15; while(buf[k]==' '||buf[k]=='\t') k++;
	  errno = 0; cur_content_length = strtoull( buf+k, &stopchar, 0 );
	  if( errno==0 && ( !*stopchar || *stopchar == '\r' || *stopchar == '\n' || *stopchar == ' ' || *stopchar == '\t' ) ) {
	    if(!had_content_length) {
	      content_length = dld_pos + cur_content_length;
	    } else if ( dld_pos + cur_content_length != content_length ) { 
	      fprintf(stderr,"file size has changed at restart of download; interrupting download.\n");
	      errcode=FileChangedErr; break2=true; break3=true; break;
	    }
	    had_content_length = true;
	  } else {
	    fprintf(stderr,"error parsing Content-Length http header field field\n");
	    errcode=HttpErr; break2=true; break3=true; break;
	  }
	} else if( !strncasecmp(buf+i,"Location:",min_macro(9,j-i)) && ( !header_continued || i>0 ) ) {
	  l=i+9; while( l<j && buf[l]!='\r' && buf[l]!='\n' ) l++;
	  if( !printheaders ) { fwrite(buf+i,l-i,1,stderr); fprintf(stderr,"\n\n"); }
	}
	i=j+1;   // buf[j]=='\n', i<=len
      } while(i<len);
      if(break2) break;
      if( i > 0 ) {
	if(printheaders) fwrite(buf,i,1,hdrout);
	memmove(buf,buf+i,len-i);
	len=len-i; bufptr=buf+len; header_continued=false; 
      } else if( j >= BUFSZ ) {
	if(printheaders) fwrite(buf,BUFSZ,1,hdrout);
	bufptr=buf; len=0; header_continued=true;  // this is an error condition: line does not fit into buf
      }
    }
    // end: read and parse response header
    if(make_retry) { wasInterrupted=true; continue; }
    if(break3) break;
    if( printheaders && j>0 ) fwrite(buf,j,1,hdrout);
    if( !printheaders && verbosity >= 2 && had_content_length ) {
      fprintf(stderr,"Content-Length: %llu\n", content_length );
    }
    if( statuscode == 416 ) {  // partial content
      pthread_mutex_lock(&prnProgressMtx);
       viewProgress = false;
      pthread_mutex_unlock(&prnProgressMtx);
      if(progressThreadStarted) {
	pthread_cond_signal(&shallQuitProgress);
	pthread_join(progressThread,NULL);
	progressThreadStarted = false;
      }
      if( dld_pos > 0 && ( is_continued || wasInterrupted ) ) {  // Range: bytes=0- does not appear to produce that error
	statuscode = 200; errcode=0; fprintf(stderr,"apparently already fully downloaded: %s\n",urlspec->fullURL); 
	saveOutput = false; saveOutput_set = true;
      } else fprintf(stderr,"416 Requested Range Not Satisfiable\n");  // normally we don´t come here
    }
    //fprintf(stdout,"<%llu,%i>\n",content_length,had_content_length); printf("[%i,%i]\n",j,len);
    if(!saveOutput_set) {
      saveOutput = statuscode == 200 || statuscode == 206 || ( saveErrorResponse && !is_continued );
      if( saveOutput && !is_continued && !output ) { 
	if(!OpenOutput(urlspec)) { errcode = FileErr; break; }
    } }
    // save beginning of content to file/ouptut
    if( j < len ) {
      if(saveOutput) fwrite(buf+j,len-j,1,output);
      dld_pos += len-j; 
      if(!quiet) UpdateProgress(dld_pos,content_length);
    }
    // main download loop
    while( !had_content_length || dld_pos < content_length ) {
      ioc = readin(fd,buf,min_macro(BUFSZ,content_length));
      if(ioc<0) { 
	if(!had_timeout && verbosity >= 2 ) { 
	  if(isProgrVisible) fputc('\n',stderr); if(wasa_sighup) fprintf(stderr,HUPMSG); else fprintf(stderr,"read error (retry #%i)\n",retry); 
	  dld_last_drawn_pos = -1; 
        } 
	had_timeout = false; wasa_sighup = false;
	if( ConnErr > errcode ) errcode=ConnErr; 
	make_retry=true; break; 
      }
      if(!ioc) { 
	if( had_content_length && dld_pos < content_length ) { 
	  if(isProgrVisible) fputc('\n',stderr); fprintf(stderr,"short read: connection closed before end of file has been reached\n"); 
	  if( ConnErr > errcode ) errcode=ConnErr; dld_last_drawn_pos = -1; break; 
	}
	break;
      }
      if(saveOutput) fwrite(buf,ioc,1,output);
      dld_pos += ioc; 
      retry = 1; if(errcode==ConnErr||errcode==CertErr) errcode = 0;  // have already downloaded something: reset retry
      if(!quiet) UpdateProgress(dld_pos,content_length);
    }
    if(make_retry) { wasInterrupted=true; continue; }
    break; // downloaded correctly, no further retry necessary
  }
  pthread_mutex_lock(&prnProgressMtx);
   viewProgress = false;
  pthread_mutex_unlock(&prnProgressMtx);
  if(progressThreadStarted) {
    pthread_cond_signal(&shallQuitProgress);
    pthread_join(progressThread,NULL);
    progressThreadStarted = false;
  }
  if(isProgrVisible) fputc('\n',stderr);
  dld_last_drawn_pos = -1;   // isProgrVisible turns false
  if( retry > max_retries && max_retries >= 0 && errcode != 0 ) {
    fprintf(stderr,"maximum number of retries exceeded!\n");
  } else {
    if( errcode==0 && ( saveOutput || saveOutput_set ) ) RenameOutput();
  }
  if(dldFileName) { fclose(output); free(dldFileName); dldFileName=NULL; }
  if(!had_content_length) {
    if(ssl) { SSL_shutdown(ssl); SSL_free(ssl); ssl = NULL; }
    if(sd>=0) { close(sd); sd=-1; }
  }
  return errcode;
}


#ifdef __USE_TIME_BITS64
  #error "hugo"
#endif

enum Command { CM_TII = 1, CM_TUU = 2, CM_TII_CERT = 3, CM_NOOP = 4 } command;
enum Mode { MD_DEFAULT = 0, MD_DANE, MD_UNSECURE_DANE, MD_CERT, MD_NO_CHECK_CERT } mode = MD_DEFAULT;
bool faaite_cert = false, check_time = true;
bool force_ipv6 = false, save_rogue_cert = true; const char *certfile;
struct UrlSpec urlspec, prev_urlspec;

bool isdldcmd(int command) { return command==CM_TII || command==CM_TUU; }
#define checkdldcmd if(!isdldcmd(command)) goto opterr

int main(int argc, char *argv[]) {
  int i, errnum; int result = 0, curres; bool next_is_always_filename = false, opterr = false; bool resolver_inited=false;
  X509 *conn_cert = NULL; void *tlsa_rec; bool reusing_connection;
  // struct stat statbuf; aux_assert(sizeof(statbuf.st_size)==8);  // correct handling of files>4GB on 32bit systems
  hdrout = stderr;

  if( argc <= 1 ) usage(true);
  if(!strcmp(argv[1],"--help")) usage(false);
  i=1;
  while( argv[i][0] == '-' ) { 
    if(!strcmp(argv[i],"--")) { i++; break; }
    if(!strcmp(argv[i],"--nsfile")) { 
      if(argc<=i+1) { fprintf(stderr,"--nsfile: missing argument\n"); return 1; }
      nsfile = argv[i+1]; i++; 
    } else if(!strcmp(argv[i],"--sys-keyfile")) { keyfile_mode |= KF_SYS; last_keyfile_mode = KF_SYS; }
    else if(!strcmp(argv[i],"--ub-keyfile")) { keyfile_mode |= KF_UNBOUND; last_keyfile_mode = KF_UNBOUND; }
    else if(!strcmp(argv[i],"--keyfile")) { 
      if(argc<=i+1) { fprintf(stderr,"--keyfile: missing argument\n"); return 1; }
      switch(last_keyfile_mode) {
	case KF_SYS: sys_keyfile = argv[i+1]; break;
	case KF_UNBOUND:unbound_keyfile = argv[i+1]; break;
	default: fprintf(stderr,"unknown keyfile mode; exiting"); return 1;
      }
      i++;
    } else { fprintf(stderr,"unknown global option: %s\n",argv[i]); opterr=true; }
    i++;
  }
  if(!strcmp(argv[i],"tii")) command = CM_TII;
  else if(!strcmp(argv[i],"tuu")) command = CM_TUU;
  else if(!strcmp(argv[i],"tii-cert")) command = CM_TII_CERT;
  else if(!strcmp(argv[i],"noop")||!strcmp(argv[i],"nop")) command = CM_NOOP;
  else { fprintf(stderr,"unknown command: %s\n\n",argv[i]); exit(1); }
  if( command == CM_TUU ) { fprintf(stderr,"not implemented: tuu(put)\n\n"); return 1; }
  i++;

  urlspec.host = NULL; prev_urlspec.host = NULL;

  netwcomm_init();
  // if(keyfile_mode) { InitResolver(); resolver_inited = true; }     // can cause internal libunbound errors at this stage
  signal(SIGWINCH,handleSigWinch); // == SIG_ERR ) fprintf(stderr,"failed to install handler for SIGWINCH.\n");
  if( signal(SIGHUP,handleSigHUP) == SIG_ERR ) fprintf(stderr,"failed to install handler for SIGHUP.\n");
  if( signal(SIGPIPE,SIG_IGN) == SIG_ERR ) fprintf(stderr,"failed to ignore SIGPIPE - may quit on restarting a https download by accident.\n");;

  while( argc > i ) {

    if( argv[i][0] == '-' && !next_is_always_filename ) {
      if(argv[i][1] != '-' ) {
	// short option
	const char *oc;
	for( oc=argv[i]+1; *oc; oc++ )
	  switch(*oc) {
	    case 'c': do_continue = true; checkdldcmd; break;
	    case 'q': quiet = true; if(verbosity>0) verbosity--; break;
	    case 'v': verbosity++; break;
	    case 'd': mode = MD_DANE; break;
	    //case 'u': mode = MD_UNSECURE_DANE; break;   // reserve this flag for user
	    case 'n': mode = MD_NO_CHECK_CERT; break;
	    case 's': fileNameMode = FS_STDOUT; break;
	    case '0': mode = MD_DEFAULT; break;
	    case '6': force_ipv6 = true; break;
	    case 'h': printheaders++; checkdldcmd;                   case 'H': printheaders = 0; checkdldcmd; break;
	    case 'o': overwriteTargetFile = true; checkdldcmd;       case 'O': overwriteTargetFile = false; checkdldcmd; break;
	    case 'e': saveErrorResponse = true; checkdldcmd; break;  case 'E': saveErrorResponse = false; checkdldcmd; break;
	    case 'i': useSNI = true; break;                case 'I': useSNI = false; break;
	    opterr: default: fprintf(stderr,"unknown option / option not valid for this command: -%c\n",*oc); opterr=true; break;
	  }
      } else {
	// long option
	if(!strcmp(argv[i],"--")) next_is_always_filename = true;
	else if(!strcmp(argv[i],"--faaite-cert")||!strcmp(argv[i],"--show-cert")) faaite_cert = true;
	else if(!strcmp(argv[i],"--faaite-ore-cert")||!strcmp(argv[i],"--no-show-cert")) faaite_cert = false;
	else if(!strcmp(argv[i],"--check-time")) check_time = true;
	else if(!strcmp(argv[i],"--no-check-time")) check_time = false;
	else if(!strcmp(argv[i],"--save-rogue-cert")) save_rogue_cert = true;
	else if(!strcmp(argv[i],"--no-save-rogue-cert")) save_rogue_cert = false;
	else if(!strcmp(argv[i],"--cert")) {
	  if(argc<=i+1) { fprintf(stderr,"--cert: missing argument\n"); return 1; }
	  if( command!=CM_TII_CERT && command!= CM_NOOP) 
	    mode = MD_CERT; 
	  else if(!faaite_cert) {
	    fprintf(stderr,"TII-CERT/NOOP --cert option given but no --faaite-cert/--show-cert: makes no sense, tii-cert&noop ignores --cert\n"); 
	    opterr = 1;
	  }
	  certfile = argv[i+1]; i++;
	  if(required_cert) X509_free(required_cert);
	  required_cert = read_cert_from_file(certfile);
	  if(faaite_cert) ShowCert(required_cert,"command line file",stdout);
	  if( check_time && !CheckCertDate(required_cert,"supplied cert: ",stderr) ) {  X509_free(required_cert); required_cert=NULL; }
	} else if(!strcmp(argv[i],"--dane")) mode = MD_DANE;
	else if(!strcmp(argv[i],"--unsecure-dane")) mode = MD_UNSECURE_DANE;
	else if(!strcmp(argv[i],"--no-check-cert")) mode = MD_NO_CHECK_CERT;
	else if(!strcmp(argv[i],"--to-stdout")) fileNameMode = FS_STDOUT;
	else if(!strcmp(argv[i],"--help")) usage(false);
	else if(!strcmp(argv[i],"--timeout")) { int64 value; char *stopchar; checkdldcmd;
	  if(argc<=i+1) { fprintf(stderr,"--timeout: missing argument\n"); return 1; }
	  errno=0; value = strtoll( argv[i+1], &stopchar, 0 );
	  if( errno!=0 || *stopchar ) { fprintf(stderr,"invalid number for --timeout\n"); return 1; }
	  timeout = value; i++;
	} else if(!strcmp(argv[i],"--retries")) { long value; char *stopchar; checkdldcmd;
	  if(argc<=i+1) { fprintf(stderr,"--retries: missing argument\n"); return 1; }
	  errno=0; value = strtol( argv[i+1], &stopchar, 0 );
	  if( errno!=0 || *stopchar || value != (int)value ) { fprintf(stderr,"invalid number for --retries\n"); return 1; }
	  max_retries = value; i++;
	} else if(!strcmp(argv[i],"--to")) {
	  if(argc<=i+1) { fprintf(stderr,"--to: missing argument\n"); return 1; }
	  fileNameMode = FS_EXPLICIT;
	  fileName = argv[i+1]; i++;
	} else { fprintf(stderr,"unknown URL option: %s\n",argv[i]); opterr=true; }
      }
    } else {
      // argv[i] does not start with '-': not an option, an URL
      if(opterr) return 1;  // { putchar('!'); 
      free(prev_urlspec.host); prev_urlspec = urlspec;
      if(parse_urlspec( argv[i], &urlspec, force_ipv6 )) return 1;
      if(verbosity>=3) print_urlspec(stdout,&urlspec);
      if( ( command == CM_TII_CERT || command == CM_NOOP ) && ( urlspec.protocol == PR_HTTP || urlspec.protocol == PR_FTP ) ) {
	fprintf(stderr,"only https and ftps use certificates and encryption: %s",urlspec.fullURL);
	continue;
      }
      if( isdldcmd(command) && !quiet && !progressThreadStarted ) {
	viewProgress=true;
	if(!pthread_create(&progressThread,NULL,ProgressThreadExec,NULL)) progressThreadStarted=true;
      }
      reusing_connection = false;
      if( command == CM_TII_CERT || command == CM_NOOP ) { reusing_connection = false; errnum = 0; }
      else switch(urlspec.protocol) {
	     case PR_HTTP: 
	       if( sd>=0 && prev_urlspec.host && !strcmp(prev_urlspec.host,urlspec.host) && prev_urlspec.port_override == urlspec.port_override ) { 
		 reusing_connection = true; errnum = 0; }
	      break;
	     case PR_HTTPS:
	       if( sd>=0 && ssl && !( SSL_get_shutdown(ssl) & ( SSL_RECEIVED_SHUTDOWN | SSL_SENT_SHUTDOWN ) ) &&
		   prev_urlspec.host && !strcmp(prev_urlspec.host,urlspec.host) && prev_urlspec.port_override == urlspec.port_override ) { 
		 reusing_connection = true; errnum = 0; }
	      break;
	   }
      if(!reusing_connection) {
	if(ssl) { SSL_shutdown(ssl); /* SSL_free(ssl);*/ ssl=NULL; }
	if(conn_cert) { X509_free(conn_cert); conn_cert=NULL; }
	if(sd>=0) close(sd);
	sd = socket(urlspec.addr->ai_addr->sa_family, SOCK_STREAM, 0);
	errnum = socket_connect(&sd,urlspec.addr,0); 
      }
      //printf("reusing connection: %i\n",reusing_connection);
      if(!errnum) {
	switch(urlspec.protocol) {
	  case PR_HTTP:
	    if( mode == MD_CERT || mode == MD_DANE || mode == MD_UNSECURE_DANE ) { fprintf(stderr,"http does not support encryption, certificates and DANE - not downloading."); break; }
	    curres = mk_HTTP_Request(&urlspec,&sd,directRead,directWrite,NULL);
	    if( curres && curres > result ) result = curres;
	    break;
	  case PR_HTTPS:
	    if( mode == MD_CERT && !required_cert ) break;  // cert supplied with --cert had invalid date: do not download
	    if(reusing_connection) {
	      curres = mk_HTTP_Request(&urlspec,ssl,sslRead,sslWrite,conn_cert); 
	      if( curres && curres > result ) result = curres;
	    } else {
	      if(!ctx_inited) { 
		SSL_library_init();
		ctx_inited = InitCTX(&ctx);
	      }
	      if(ctx_inited) {
		ssl = SSL_new(ctx);      /* create new SSL connection state */
		SSL_set_fd(ssl,sd);    /* attach the socket descriptor */
		if(useSNI) SSL_set_tlsext_host_name(ssl,urlspec.host);
		if ( SSL_connect(ssl) == -1 ) { ERR_print_errors_fp(stderr); if(result < ConnErr) result = ConnErr; }
		else { 
		  // get conn_cert and check it

		  if( isdldcmd(command) ) {
		    // tii, tuu
		    if( mode == MD_CERT ) {
		      conn_cert = SSL_get_peer_certificate(ssl);   /* get the server's certificate */
		      if(!conn_cert) { fprintf(stderr,"failed to retrieve server certificate; closing connection (%s).\n",argv[i]); if(result<CertErr) result=CertErr; break; }
		      if(!X509_compare_equal(conn_cert,required_cert)) { 
			fprintf(stderr,"server certificate does not match the required certificate; closing connection.\n"); 
			if(faaite_cert) ShowCert(conn_cert,"differing server",stdout);
			if(save_rogue_cert) saveCert(urlspec.host,"-no-match",conn_cert);
			if(result < CertErr) result = CertErr;
			break; 
		      }
		      X509_free(conn_cert);
		    } else if( mode == MD_DANE || mode == MD_UNSECURE_DANE || mode == MD_DEFAULT ) {
		      if(!resolver_inited) { InitResolver(); resolver_inited = true; }
		      if( !retrieve_TLSA( urlspec.host, &tlsa_rec, &urlspec ) && mode!=MD_UNSECURE_DANE ) { 
			if(faaite_cert) ShowCert(conn_cert,"unverified server",stdout);
			if(result<CertErr) result=CertErr; 
			break; 
		      }
		      if(!tlsa_rec) { if(result<CertErr) result=CertErr; break; }
		      conn_cert = SSL_get_peer_certificate(ssl);   /* get the server's certificate */
		      if(!verify_hash( tlsa_rec, conn_cert, &urlspec )) {
			if(faaite_cert) ShowCert(conn_cert,"rogue server",stdout);
			if(save_rogue_cert) saveCert(urlspec.host,"-rogue",conn_cert);
			if(result < CertErr) result = CertErr;
			break;
		      }
		    } else conn_cert = NULL;  // should already be NULL, but assert that here
		    if( faaite_cert && conn_cert && mode != MD_CERT ) ShowCert(conn_cert,"server",stdout);
		    if( mode != MD_NO_CHECK_CERT && check_time && !CheckCertDate(conn_cert,"server cert: ",stderr) ) { if(result<CertErr) result=CertErr; break; }
		    // start operation: download/save cert
		    curres = mk_HTTP_Request(&urlspec,ssl,sslRead,sslWrite,conn_cert); 
		    if( curres && curres > result ) result = curres;

		  } else {
		    // tii-cert, noop
		    conn_cert = SSL_get_peer_certificate(ssl);   /* get the server's certificate */
		    if(faaite_cert) ShowCert(conn_cert,"server",stdout);
		    if(!conn_cert) { fprintf(stderr,"failed to retrieve server certificate; closing connection (%s).\n",argv[i]); if(result<CertErr) result=CertErr; break; }
		    if( check_time ) CheckCertDate(conn_cert,"no consequences, server cert: ",stderr);
		    if( mode == MD_NO_CHECK_CERT) {
		      if(command!=CM_NOOP) saveCert(urlspec.host,"-not-validated",conn_cert); 
		      break; // switch
		    }
		    if(!resolver_inited) { InitResolver(); resolver_inited = true; }
		    curres = retrieve_TLSA( urlspec.host, &tlsa_rec, &urlspec );
		    if( tlsa_rec ) {
		      if( !curres || !verify_hash( tlsa_rec, conn_cert, &urlspec ) ) { 
			if( save_rogue_cert && command!=CM_NOOP ) saveCert(urlspec.host,curres?"-rogue":"-no-DNSSEC",conn_cert);
			if(result<2) result=2; 
		      } else if(command!=CM_NOOP) saveCert(urlspec.host,"",conn_cert);
		    } else { 
		      if( save_rogue_cert && command!=CM_NOOP ) saveCert(urlspec.host,"-not-validated",conn_cert); 
		      if(result<2) result=2; 
		  } } 

	    } } }
	    break;
	}
      } else if(result > ConnErr) curres = ConnErr;  // errnum was set from socket_connect
      fileNameMode = FS_IMPLICIT;
      next_is_always_filename = false;
      // end: argv[i] does not start with '-': not an option, an URL
    }

    i++;
  } 
  //if(ssl) { SSL_shutdown(ssl); SSL_free(ssl); ssl=NULL; }
  if(conn_cert) { X509_free(conn_cert); conn_cert=NULL; }
  if(sd>=0) close(sd);
  // if(ctx_inited) SSL_CTX_free(ctx);     // sometimes segfaults
  if(required_cert) X509_free(required_cert);
  if(resolver_inited) CleanUpResolver();
  free(urlspec.host); free(prev_urlspec.host);
  netwcomm_cleanup();
  return result;
}
