
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <arpa/inet.h> 
//#include <resolv.h>
#include <netdb.h>

#include <sys/types.h> 
#include <sys/stat.h> 

#include <unistd.h>
#include <alloca.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

#include "networking.h"
#include "dane-support.h"
// --

//#define SYS_KEYFILE "/usr/share/dns/root.key"
//#define SYS_KEYFILE_2 "/etc/trusted-key.key"
//#define UB_KEYFILE "/etc/unbound/root.key"

//#define DF_NSFILE_1 "/etc/resolv.conf.DANE"
//#define DF_NSFILE_2 "/etc/resolv.conf"

//enum NSFileMode { NS_DEFAULT, NS_EXPLICIT };
enum NSFileMode nsfile_mode = NS_DEFAULT;
const char *nsfile = NULL;   // for NS_EXPLICIT only

// enum KeyFileMode { KF_DEFAULT = 0, KF_UNBOUND = 1, KF_SYS = 2 };   // default is effectively the same as unbound
enum KeyFileMode keyfile_mode = KF_DEFAULT;  // set to zero, bit array
enum KeyFileMode last_keyfile_mode = KF_DEFAULT;  // used to see of which type a keyfile stated as parameter will be
const char *unbound_keyfile = UB_KEYFILE, *sys_keyfile = SYS_KEYFILE; 

struct __res_state rctx;

bool InitResolver() {
  // const char *eff_nsfile; int retval; struct stat statbuf;
  if(res_ninit(&rctx)) { fprintf(stderr,"error initializing DNS resolver with res_ninit.\n"); return false; }
  /*
  if( keyfile_mode == KF_DEFAULT ) {
    if(stat(sys_keyfile,&statbuf)==0) keyfile_mode = KF_SYS;
    else if(stat(unbound_keyfile,&statbuf)==0) keyfile_mode = KF_UNBOUND; 
    else { fprintf(stderr,"DANE: keyfile not found (neither '%s' nor '%s')\n",sys_keyfile,unbound_keyfile); return false; }
  }
  if( keyfile_mode & KF_SYS )
    if( ( retval = ub_ctx_add_ta_file(rctx, sys_keyfile) ) != 0 ) { fprintf(stderr,"DANE: error adding public DNSSEC keys: %s; errno says: %s\n", ub_strerror(retval),strerror(errno)); return false; }
  if( keyfile_mode & KF_UNBOUND )
    if( ( retval = ub_ctx_trustedkeys(rctx, unbound_keyfile) ) != 0 ) { fprintf(stderr,"DANE: error adding public DNSSEC keys: %s; errno says: %s\n", ub_strerror(retval),strerror(errno)); return false; }
  */
  return true;
}

#define ANSLEN 8192
struct Answer {
  int len;
  u_char content[NS_PACKETSZ];
  //unsigned char content[ANSLEN];
} answer;

bool retrieve_TLSA( const char *domain, void **tlsa_rec, struct UrlSpec *urlspec ) {
  int len = strlen(domain); int bufsz; char *domainIdent = alloca(bufsz=len+13);
  int lenret;
  snprintf( domainIdent, bufsz, "_%i._tcp.%s.", urlspec->port_override>=0 ? urlspec->port_override : 443, domain );
 //retval = res_nquerydomain( rctx, "_443._tcp.", domain, C_IN, T_TLSA, answer.content, ANSLEN );
  //retval = res_nsearch( rctx, "_443._tcp.elstel.org.", C_IN, T_TLSA, answer.content, ANSLEN );
  lenret = res_nquery( &rctx, domainIdent, ns_c_in, ns_t_tlsa, answer.content, sizeof(answer.content) );
  while(lenret<0) {
    domain = strchr(domain,'.');
    if(!domain++||!*domain) break;
    snprintf( domainIdent, bufsz, "_%i._tcp.%s.", urlspec->port_override>=0 ? urlspec->port_override : 443, domain );
    lenret = res_nquery( &rctx, domainIdent, ns_c_in, ns_t_tlsa, answer.content, sizeof(answer.content) );
  }
  if( lenret >=0 ) {
    answer.len = lenret;
    *tlsa_rec = &answer;
  } else { 
    herror("DANE: could not retrieve TLSA record");
    //fprintf(stderr,"DANE: could not retrieve TLSA record: %s\n",hstrerror(h_errno));
    *tlsa_rec = NULL; return false; 
  }
  return true;
}

bool verify_hash( void *answer_asvoid, X509 *cert, struct UrlSpec *urlspec ) {
  int l, i, count; const uchar *thisdata; bool result = false, retval; 
  int lenA; uchar *bufA; X509_PUBKEY *pubkey;
  struct Answer *answer = (struct Answer*)answer_asvoid; 
  ns_msg handle; ns_rr rr; uchar digest[64];  // enough for SHA512
  ns_initparse(answer->content, answer->len, &handle);
  count = ns_msg_count(handle,ns_s_an);
  for(i=0;i<count;i++)
    if(ns_parserr(&handle, ns_s_an, i, &rr) == 0) {   // undocumented: it is questionable whether third param should be i or zero
      l = ns_rr_rdlen(rr); thisdata = ns_rr_rdata(rr); 
      // result len=35; 03:00:01:a8:ed:f0:ca:ca:f7:76:ac:ac:df:e5:35:64:c5:15:56:ad:32:5f:03:a3:69:e4:c8:f4:62:2b:4d:c5:b0:68:65:,
      //for( i = 0; i < l; i++ ) printf("%02x:",(int)(unsigned char)(*thisdata)[i]);  // ignore three bytes
      if( thisdata[2] != ST_NONE ) {
	retval = sha_of_cert( digest, thisdata[1], thisdata[2], cert, urlspec->fullURL );
	if( l != shalen[(int)thisdata[2]] + 3 ) fprintf(stderr,"DANE: warning: Length of TLSA record not as expected (=%i)! It may contain invalid information (%s).\n",l,urlspec->fullURL);
	if(!retval||shalen[(int)thisdata[2]]>l-3) continue;
	if(!memcmp(thisdata+3,digest,shalen[(int)thisdata[2]])) { result = true; break; }
	if( !result && verbosity>=2 ) {
	  fprintf(stderr,"TLSA record (first three bytes are for TLSA-mode): ");
	  for( i = 0; i < l; i++ ) fprintf(stderr,"%s%02x",i?":":"",(int)(unsigned char)(thisdata)[i]); fputc('\n',stderr);   // ignore three bytes
	  fprintf(stderr,"hash of the server certificate: ");
	  for( i = 0; i < shalen[(int)(thisdata)[2]]; i++ ) fprintf(stderr,"%s%02x",i?":":"",(int)(unsigned char)digest[i]); fputc('\n',stderr);  // ignore three bytes
	}
      } else {
	switch((thisdata)[1]) {
	  case SK_FULLCERT:
	    lenA = i2d_X509(cert,&bufA);
	    if( lenA == l - 3  && !memcmp(thisdata+3,bufA,lenA) ) result = true;
	    OPENSSL_free(bufA);
	    break;
	  case SK_PUBKEY:
	    pubkey = X509_get_X509_PUBKEY(cert);
	    lenA = i2d_X509_PUBKEY(pubkey,&bufA);
	    X509_PUBKEY_free(pubkey);
	    if( lenA == l - 3  && !memcmp(thisdata+3,bufA,lenA) ) result = true;
	    OPENSSL_free(bufA);
	    break;
	  default:
	    fprintf(stderr,"DANE: verify_hash: unknown ShaWhat mode (%s)\n",urlspec->fullURL);
	    break;
	}
	if(result) break;
      }
    } 
  if(!result) fprintf(stderr,"certificate signature in TLSA record did not match (%s)\n",urlspec->fullURL);
  return result;
}

void CleanUpResolver() {
  res_nclose(&rctx);
}

