// for copying and licensing see the attached file LICENSE.txt; please keep this file with the software
//
#ifndef _SOCK_COMMON_H
#define _SOCK_COMMON_H
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>	  // in add. to pthread.h: EOWNERDEAD return value
#include <stdio.h>
#include <openssl/ssl.h>
#include <arpa/inet.h>
#include "basedefines.h"

#define aux_sysmsg(message) fputs(message,stderr);
#define aux_error(message) aux_sysmsg(message)
#define aux_warning(message) aux_sysmsg(message)
#define aux_sysmsg_int(value) fprintf(stderr,"%lli",(long long)value);
#define aux_sysmsg_adr(value) fprintf(stderr,"%llx",(unsigned long long)value);

#define aux_praefixed_assert(condition,praefix,message) if(unlikely(!(condition))) { aux_error(praefix "assert failed at " __FILE__ ":"); aux_sysmsg_int(__LINE__); aux_sysmsg(": " message #condition "\n\n"); abort(); }
#define aux_assert_(condition,message) aux_praefixed_assert(condition,"aux_",message " ~ ")
#define aux_assert(condition) aux_praefixed_assert(condition,"aux_","")

extern int verbosity;   // default is 1

int setmicroseconds( struct timeval *tv, const char *mstxt, long jiffie );
int setnanoseconds( struct timespec *ts, const char *mstxt, long jiffie );
// pass 0 or a negative value for jiffie to accept any timespan; otherwise jiffie gives the minimum frequency

int doclose(int fd);
void dup2quitonerr( int oldfd, int newfd );

bool checkascii( const char *str );

//
// openssl routines
//

extern BIO *membio; extern BIO *stdout_bio;
int netwcomm_init(); void netwcomm_cleanup();

enum ShaType { ST_NONE = 0, ST_SHA256 = 1, ST_SHA512 = 2, ST_SHA1 = 3, ST_count = 4 };  // ShaType, equals number in TLSA record (0,1,2); must not change!!
enum ShaWhatOfCert { SK_FULLCERT = 0, SK_PUBKEY = 1 };   // ShaKind, equals number in TLSA record
extern int shalen[ST_count];
bool sha_of_cert( uchar *digest, int shaType, int shaWhat, X509 *cert, const char *err_comment );  // prints error messages

enum PrnHexFlags { PRNHEX_NOSEP = 1, PRNHEX_NONL = 2 };
void prn_as_hex(FILE *out, const char *message, const uchar *digest, int len, int flags );
int X509name2utf( BIO *bio, X509_NAME *nm, int indent );
bool print_v3_ext( FILE *out, X509 *cert, int nid, const char *tagname, int indent );
void ShowCert(X509 *cert,const char *whatacert,FILE *out);
bool CheckCertDate( X509 *cert, const char *msghdr, FILE *out );

X509* read_cert_from_file( const char *filename );
bool X509_compare_equal( X509 *certA, X509 *certB );

typedef int(*PWriteOut)(void*,const void *,int);
typedef int(*PReadIn)(void*,void*,int);
extern uint64 pn;  // counter intialized to 0 and incremented for debugging purposes

int directWrite( void *fd, const void *buf, int len );
int directRead( void *fd, void *buf, int len );

int sslWrite( void *fd, const void *buf, int len );
int sslRead( void *fd, void *buf, int len );

//
// general socket routines
//

// priamry sending, receival and temporary buffer
#define SF_ANYHOST 1
#define SF_IPV6 2
#define SF_SILENT 4

enum Protocol { PR_HTTP = 1, PR_HTTPS = 2, PR_FTP = 3, PR_FTPS = 4 };
extern const char* protocol2str[3];

struct UrlSpec {
  struct addrinfo *addr; int protocol, port_override;  // flags: 1~anyhost, 2~ipv6
  char *host; const char *path, *target, *fullURL; bool isip6adr;
};

// result==0: not an ipv6, result==-1: looked like an ipv6; printed error, result>0: host, in6addr filled correctly
// result==1: ok, result==2: output a warning because of formatting error, *port may be written if a port override was given otherwise it is not touched
int parse_ipv6( char *start, char **host, struct in6_addr *in6addr, uint16_t *port, int *out_k );

int parse_urlspec( char *url, struct UrlSpec *urlspec, bool force_ipv6 );
void print_sockaddr( FILE* out, struct sockaddr *sa );
void print_urlspec( FILE* out, struct UrlSpec *urlspec );

// flags: SF_ANYHOST illegal; SF_IPV6 socket_connect ignores it an expects ipv6-forcing in addrinfo
// if signal handler set *socket=-2 or a value below then socket_connect returns EAGAIN
int socket_connect( int *socket, struct addrinfo *listening_host_ais, int flags ); 

#if OSTYPE & OS_linux
  #define OSSTR " (linux-gnu)"
#elif OSTYPE & OS_OpenBSD
  #define OSSTR " (OpenBSD)"
#elif OSTYPE & OS_BSD
  #define OSSTR " (BSD)"
#elif OSTYPE & OS_windows
  #define OSSTR " (Windows)"
#elif OSTYPE & OS_macos
  #define OSSTR " (MacOS)"
#elif OSTYPE & OS_os2
  #define OSSTR " (OS/2)"
#elif OSTYPE & OS_unix
  #define OSSTR " (Unix)"
#elif OSTYPE & OS_dos
  #define OSSTR " (DOS)"
#else
  #define OSSTR ""
#endif


#endif
