// for copying and licensing see the attached file LICENSE.txt; please keep this file with the software
//
#include <alloca.h>
#include <openssl/dsa.h>
//#include <openssl/pem.h>
#include <openssl/asn1.h>
#include "networking.h"
#include "sslcrypto.h"

bool parse_pubkey_d( struct pubkey *out,  const uchar *asn1key, int len ) {
  int i; bool result, allgood, has_pqg; const uchar *keyptr; ASN1_SEQUENCE_ANY *seq, *seq2, *seq3; 
  ASN1_STRING *asn1str; ASN1_OBJECT *obj; ASN1_INTEGER *modulo, *exponent; // ASN1_BIT_STRING *bitstr; 
  out->flags = 0; result=false; allgood=true; seq3=NULL;
  keyptr=asn1key; seq = d2i_ASN1_SEQUENCE_ANY(NULL,&keyptr,len);
  //prn_as_hex(stdout,"full pristine key: ",asn1key, len, PRNHEX_NOSEP );
  if( sk_ASN1_TYPE_num(seq) != 2 ) goto freeseq0;
  if( V_ASN1_SEQUENCE != ASN1_TYPE_get(sk_ASN1_TYPE_value(seq,0)) ) goto freeseq0;
  asn1str = sk_ASN1_TYPE_value(seq,0)->value.sequence;
  keyptr = ASN1_STRING_get0_data(asn1str);
  seq2 = d2i_ASN1_SEQUENCE_ANY( NULL, &keyptr, ASN1_STRING_length(asn1str) );
  if( sk_ASN1_TYPE_num(seq2) < 2 ) goto freeseq;   // normally 2, at least for DSA & RSA (including NULL-obj padding)
  if( V_ASN1_OBJECT != ASN1_TYPE_get(sk_ASN1_TYPE_value(seq2,0)) ) goto freeseq;
  obj=sk_ASN1_TYPE_value(seq2,0)->value.object;
  out->type_nid = OBJ_obj2nid(obj); out->flags|=PBK_HAS_TYPE; result=true;
  //OBJ_obj2txt(buff, 1024, obj, 0); buff[1023]='\000';
  allgood = allgood && ( sk_ASN1_TYPE_num(seq2) == 2 );
  if( V_ASN1_BIT_STRING != ASN1_TYPE_get(sk_ASN1_TYPE_value(seq,1)) ) goto freeseq;
  asn1str = sk_ASN1_TYPE_value(seq,1)->value.bit_string;
  keyptr = ASN1_STRING_get0_data(asn1str);

  if( out->type_nid == NID_rsaEncryption ) {
    allgood = allgood && ( V_ASN1_NULL == ASN1_TYPE_get(sk_ASN1_TYPE_value(seq2,1)));
    seq3 = d2i_ASN1_SEQUENCE_ANY(NULL,&keyptr,ASN1_STRING_length(asn1str));
    if( sk_ASN1_TYPE_num(seq3) != 2 || \
        V_ASN1_INTEGER != ASN1_TYPE_get(sk_ASN1_TYPE_value(seq3,0)) || \
        V_ASN1_INTEGER != ASN1_TYPE_get(sk_ASN1_TYPE_value(seq3,1))        ) { allgood=false; goto freeseq3; }
    modulo = sk_ASN1_TYPE_value(seq3,0)->value.integer; exponent = sk_ASN1_TYPE_value(seq3,1)->value.integer;
    if( ASN1_INTEGER_get_int64((int64_t*)&out->fields.rsa.exponent,exponent) ) out->flags|=PBK_HAS_exponent; else allgood=false;
    out->fields.rsa.modulo = ASN1_INTEGER_to_BN(modulo,NULL);
    if(out->fields.rsa.modulo!=NULL) out->flags|=PBK_HAS_modulo|PBK_KNOW_KEYSIZE; else allgood=false;

  } else if( out->type_nid == NID_dsa ) {  // NID_dsa_2, NID_dsaWithSHA etc.
    ASN1_INTEGER *asn1int;
    asn1int = d2i_ASN1_INTEGER(NULL,&keyptr,ASN1_STRING_length(asn1str));
    if(asn1int) {
      out->fields.dsa.y = ASN1_INTEGER_to_BN(asn1int,NULL);
      out->flags|=PBK_HAS_Y;
      ASN1_INTEGER_free(asn1int);
    } else allgood=false; 
    if( V_ASN1_SEQUENCE != ASN1_TYPE_get(sk_ASN1_TYPE_value(seq2,1)) ) { result=false; goto freeseq; }
    asn1str = sk_ASN1_TYPE_value(seq2,1)->value.sequence;
    keyptr = ASN1_STRING_get0_data(asn1str);
    seq3 = d2i_ASN1_SEQUENCE_ANY(NULL,&keyptr,ASN1_STRING_length(asn1str));
    if( sk_ASN1_TYPE_num(seq3) < 3 ) { result=false; goto freeseq3; }
    allgood = allgood && sk_ASN1_TYPE_num(seq3) == 3;
    for(has_pqg=true,i=0;i<3;i++) has_pqg = has_pqg && V_ASN1_INTEGER == ASN1_TYPE_get(sk_ASN1_TYPE_value(seq3,i)); 
    if(has_pqg) {
      out->fields.dsa.p = ASN1_INTEGER_to_BN(sk_ASN1_TYPE_value(seq3,0)->value.integer,NULL);
      out->fields.dsa.q = ASN1_INTEGER_to_BN(sk_ASN1_TYPE_value(seq3,1)->value.integer,NULL);
      out->fields.dsa.g = ASN1_INTEGER_to_BN(sk_ASN1_TYPE_value(seq3,2)->value.integer,NULL);
      out->flags|=PBK_HAS_PQG;
    } else allgood=false;

  }
freeseq3:
  if(seq3) {
    for(i=0;i<sk_ASN1_TYPE_num(seq3);i++) ASN1_TYPE_free(sk_ASN1_TYPE_value(seq3,i));
    sk_ASN1_TYPE_free(seq3);
  }
freeseq:
  for(i=0;i<sk_ASN1_TYPE_num(seq2);i++) ASN1_TYPE_free(sk_ASN1_TYPE_value(seq2,i));
  sk_ASN1_TYPE_free(seq2);
freeseq0:
  for(i=0;i<sk_ASN1_TYPE_num(seq);i++) ASN1_TYPE_free(sk_ASN1_TYPE_value(seq,i));
  sk_ASN1_TYPE_free(seq);
  if( allgood && result ) out->flags|=PBK_ALL_GOOD;
  return result;
}

void free_pubkey( struct pubkey *pbk ) {
  switch(pbk->type_nid) {
    case NID_rsaEncryption:
      //if( pbk->flags & PBK_HAS_exponent ) BN_free(pbk->fields.rsa.exponent);
      if( pbk->flags & PBK_HAS_modulo ) BN_free(pbk->fields.rsa.modulo);
     break;
    case NID_dsa:
      if( pbk->flags & PBK_HAS_Y ) BN_free(pbk->fields.dsa.y);
      if( pbk->flags & PBK_HAS_PQG ) { BN_free(pbk->fields.dsa.p); BN_free(pbk->fields.dsa.q); BN_free(pbk->fields.dsa.g); }
     break;
  }
}

bool parse_pubkey( struct pubkey *out, EVP_PKEY *pubkey ) {
  int len; uchar *asn1key, *keyptr;
  len = i2d_PUBKEY(pubkey,NULL); asn1key = keyptr = alloca(len); len = i2d_PUBKEY(pubkey,&keyptr);
  return parse_pubkey_d( out, (const uchar*)asn1key, len );
}

EVP_PKEY* construct_pubkey_rsa(struct pubkey *pbk);
EVP_PKEY* construct_pubkey_dsa(struct pubkey *pbk);

EVP_PKEY* construct_pubkey(struct pubkey *pbk) {
  if( pbk->type_nid == NID_rsaEncryption ) return construct_pubkey_rsa(pbk);
  if( pbk->type_nid == NID_dsa ) return construct_pubkey_dsa(pbk);
  return NULL;
}

EVP_PKEY* construct_pubkey_dsa(struct pubkey *pbk) {
  ASN1_INTEGER *ap, *aq, *ag, *ay; STACK_OF(ASN1_TYPE) *paramseq, *hdrseq, *mainseq; 
  ASN1_TYPE *ip, *iq, *ig, *item_type_nid, *item_paramseq, *item_hdrseq, *item_bitstr;
  ASN1_OCTET_STRING *o_paramseq, *o_hdrseq; ASN1_BIT_STRING *bitstr;
  int res, result, datalen; uchar *data, *dataptr;
  result = true;

  ap = BN_to_ASN1_INTEGER(pbk->fields.dsa.p,NULL);
  aq = BN_to_ASN1_INTEGER(pbk->fields.dsa.q,NULL);
  ag = BN_to_ASN1_INTEGER(pbk->fields.dsa.g,NULL);
  paramseq = sk_ASN1_TYPE_new_null();
  ip = ASN1_TYPE_new(); ASN1_TYPE_set(ip,V_ASN1_INTEGER,ap);
  iq = ASN1_TYPE_new(); ASN1_TYPE_set(iq,V_ASN1_INTEGER,aq);
  ig = ASN1_TYPE_new(); ASN1_TYPE_set(ig,V_ASN1_INTEGER,ag);
  sk_ASN1_TYPE_push(paramseq,ip);
  sk_ASN1_TYPE_push(paramseq,iq);
  sk_ASN1_TYPE_push(paramseq,ig);
  datalen = i2d_ASN1_SEQUENCE_ANY(paramseq,NULL); data = dataptr = (uchar*)alloca(datalen);
  res = i2d_ASN1_SEQUENCE_ANY(paramseq,&dataptr); if(res<0) { result=false; goto free_paramseq; }
  o_paramseq = ASN1_OCTET_STRING_new();
  ASN1_OCTET_STRING_set(o_paramseq,data,datalen);

  item_type_nid = ASN1_TYPE_new(); ASN1_TYPE_set(item_type_nid,V_ASN1_OBJECT,OBJ_nid2obj(pbk->type_nid));
  item_paramseq = ASN1_TYPE_new(); ASN1_TYPE_set(item_paramseq,V_ASN1_SEQUENCE,o_paramseq);
  hdrseq = sk_ASN1_TYPE_new_null();
  sk_ASN1_TYPE_push(hdrseq,item_type_nid);
  sk_ASN1_TYPE_push(hdrseq,item_paramseq);
  datalen = i2d_ASN1_SEQUENCE_ANY(hdrseq,NULL); data = dataptr = (uchar*)alloca(datalen);
  res = i2d_ASN1_SEQUENCE_ANY(hdrseq,&dataptr); if(res<0) { result=false; goto free_hdrseq; }
  o_hdrseq = ASN1_OCTET_STRING_new();
  ASN1_OCTET_STRING_set(o_hdrseq,data,datalen);
  item_hdrseq = ASN1_TYPE_new(); ASN1_TYPE_set(item_hdrseq,V_ASN1_SEQUENCE,o_hdrseq);

  ay = BN_to_ASN1_INTEGER(pbk->fields.dsa.y,NULL);
  datalen = i2d_ASN1_INTEGER(ay,NULL); data = dataptr = (uchar*)alloca(datalen);
  res = i2d_ASN1_INTEGER(ay,&dataptr); if(res<0) { result=false; goto free_ay; }
  bitstr = ASN1_BIT_STRING_new();
  ASN1_BIT_STRING_set(bitstr,data,datalen);
  item_bitstr = ASN1_TYPE_new(); ASN1_TYPE_set(item_bitstr,V_ASN1_BIT_STRING,bitstr);

  mainseq = sk_ASN1_TYPE_new_null();
  sk_ASN1_TYPE_push(mainseq,item_hdrseq);
  sk_ASN1_TYPE_push(mainseq,item_bitstr);
  datalen = i2d_ASN1_SEQUENCE_ANY(mainseq,NULL); data = dataptr = (uchar*)alloca(datalen);
  res = i2d_ASN1_SEQUENCE_ANY(mainseq,&dataptr); if(res<0) result=false;

  sk_ASN1_TYPE_free(mainseq);     // part of 
  ASN1_TYPE_free(item_bitstr);
free_ay:
  ASN1_TYPE_free(item_hdrseq); ASN1_INTEGER_free(ay);          // parts of hdrseq
  sk_ASN1_TYPE_free(hdrseq);     // part of 
free_hdrseq:
  ASN1_TYPE_free(item_type_nid); ASN1_TYPE_free(item_paramseq);          // parts of hdrseq
  sk_ASN1_TYPE_free(paramseq);     // part of 
free_paramseq:
  ASN1_TYPE_free(ip); ASN1_TYPE_free(iq); ASN1_TYPE_free(ig);     // parts of paramseq
  if(!result) return NULL;
  //printf("rc_len=%i\n",datalen);
  return d2i_PUBKEY(NULL,(const uchar**)&data,datalen);
}

EVP_PKEY* construct_pubkey_rsa(struct pubkey *pbk) {
  ASN1_INTEGER *modulo, *exponent; STACK_OF(ASN1_TYPE) *dataseq, *hdrseq, *mainseq; ASN1_BIT_STRING *bitstr; 
  ASN1_TYPE *item_modulo, *item_exponent, *item_type_nid, *item_null, *item_hdrseq, *item_bitstr; 
  ASN1_OCTET_STRING *ocstr;
  int res, result, datalen; uchar *data, *dataptr;
  result = true;

  modulo = BN_to_ASN1_INTEGER(pbk->fields.rsa.modulo,NULL);
  exponent = ASN1_INTEGER_new();
  ASN1_INTEGER_set_int64( exponent, (int64_t)pbk->fields.rsa.exponent );
  dataseq = sk_ASN1_TYPE_new_null();
  item_modulo = ASN1_TYPE_new(); ASN1_TYPE_set(item_modulo,V_ASN1_INTEGER,modulo);
  item_exponent = ASN1_TYPE_new(); ASN1_TYPE_set(item_exponent,V_ASN1_INTEGER,exponent);
  sk_ASN1_TYPE_push(dataseq,item_modulo);
  sk_ASN1_TYPE_push(dataseq,item_exponent);
  datalen = i2d_ASN1_SEQUENCE_ANY(dataseq,NULL); data = dataptr = (uchar*)alloca(datalen);
  res = i2d_ASN1_SEQUENCE_ANY(dataseq,&dataptr); if(res<0) { result=false; goto free_dataseq; }
  bitstr = ASN1_BIT_STRING_new();
  ASN1_BIT_STRING_set(bitstr,data,datalen);

  item_type_nid = ASN1_TYPE_new(); ASN1_TYPE_set(item_type_nid,V_ASN1_OBJECT,OBJ_nid2obj(pbk->type_nid));
  item_null = ASN1_TYPE_new(); ASN1_TYPE_set(item_null,V_ASN1_NULL,ASN1_NULL_new());
  hdrseq = sk_ASN1_TYPE_new_null();
  sk_ASN1_TYPE_push(hdrseq,item_type_nid);
  sk_ASN1_TYPE_push(hdrseq,item_null);
  // that is a hack: with openssl you can not directly pack a sequence into a sequence
  // that is why we make an octet string object of the sequence and then simply use a sequence nid on it
  datalen = i2d_ASN1_SEQUENCE_ANY(hdrseq,NULL); data = dataptr = (uchar*)alloca(datalen);
  res = i2d_ASN1_SEQUENCE_ANY(hdrseq,&dataptr); if(res<0) { result=false; ASN1_BIT_STRING_free(bitstr); goto free_hdrseq; }
  ocstr = ASN1_OCTET_STRING_new();
  ASN1_OCTET_STRING_set(ocstr,data,datalen);

  item_hdrseq = ASN1_TYPE_new(); ASN1_TYPE_set(item_hdrseq,V_ASN1_SEQUENCE,ocstr);
  item_bitstr = ASN1_TYPE_new(); ASN1_TYPE_set(item_bitstr,V_ASN1_BIT_STRING,bitstr);
  mainseq = sk_ASN1_TYPE_new_null();
  sk_ASN1_TYPE_push(mainseq,item_hdrseq);
  sk_ASN1_TYPE_push(mainseq,item_bitstr);
  datalen = i2d_ASN1_SEQUENCE_ANY(mainseq,NULL); data = dataptr = (uchar*)alloca(datalen);
  res = i2d_ASN1_SEQUENCE_ANY(mainseq,&dataptr); if(res<0) result=false;
  //prn_as_hex(stdout,"key: ",data, datalen, PRNHEX_NOSEP );

  sk_ASN1_TYPE_free(mainseq);   // sk_TYPE_free only frees the sequenece, not its elements
  ASN1_TYPE_free(item_hdrseq);  // ocstr is part of mainseq, ocstr used to pack hdrseq
  ASN1_TYPE_free(item_bitstr);   // part of hdrseq
  sk_ASN1_TYPE_free(hdrseq);      // part of mainseq/ocstr, free only if already encoded good
free_hdrseq:
  ASN1_TYPE_free(item_type_nid); ASN1_TYPE_free(item_null);   // parts of hdrseq
  sk_ASN1_TYPE_free(dataseq);     // part of bitstr, free only if encoded good
free_dataseq:
  ASN1_TYPE_free(item_exponent); ASN1_TYPE_free(item_modulo);   // parts of dataseq
  //already freed: ASN1_INTEGER_free(exponent); ASN1_INTEGER_free(modulo);
  if(!result) return NULL;
  //printf("rc_len=%i\n",datalen);
  return d2i_PUBKEY(NULL,(const uchar**)&data,datalen);
}

int get_keysize(struct pubkey *pbk) {
  if( pbk->type_nid == NID_rsaEncryption ) 
    return BN_num_bits(pbk->fields.rsa.modulo);
  return 0;
}

bool get_DSA_keysize_p_q(struct pubkey *pbk, int *p, int *q ) {
  if( pbk->type_nid == NID_dsa && ( pbk->flags & PBK_HAS_PQG ) ) {
    *p = BN_num_bits(pbk->fields.dsa.p);
    *q = BN_num_bits(pbk->fields.dsa.q);
    return true;
  }
  return false;
}

